# coding: utf-8

u"""Генерация отчётов средствами платформы M3."""

import os
import sys
import uuid

from django.conf import settings

import objectpack
from m3 import ApplicationLogicException
from m3.actions import OperationResult

from simple_report.converter.abstract import FileConverter
from simple_report.report import SpreadsheetReport, DocumentReport
from simple_report.xls.document import DocumentXLS

from . import BaseReportPack


def download_result(url):
    u"""функция для скачивания файла отчёта."""
    return OperationResult(
        success=True,
        code=u"""
            (function(){
                var hiddenIFrameID = 'hiddenDownloader',
                    iframe = document.getElementById(hiddenIFrameID);
                if (iframe === null) {
                    iframe = document.createElement('iframe');
                    iframe.id = hiddenIFrameID;
                    iframe.style.display = 'none';
                    document.body.appendChild(iframe);
                }
                iframe.src = "%s";
            })()
        """ % url
    )


path_dir = settings.REPORTS_DIR
if not os.path.exists(path_dir):
    os.makedirs(path_dir)


def get_path(filename):
    """
    Путь до временного файла отчета
    :param str filename: имя файла (с раcширением, без пути до файла)
    :rtype: str
    """
    return os.path.abspath(os.path.join(path_dir, filename))


def get_url(filename):
    """
    URL до временного файла отчета
    :param str filename: имя файла (с раcширением, без пути до файла)
    :rtype: str
    """
    return '/'.join((settings.REPORTS_URL, filename))


class CommonReportWindowAction(objectpack.BaseWindowAction):

    u"""Экшн показа окна отчёта (перед выполнением отчёта)."""

    perm_code = 'report'

    def create_window(self):
        self.win = self.parent.create_report_window(self.request, self.context)

    def configure_window(self):
        self.win.save_btn.text = u'Сформировать'

    def set_window_params(self):
        super(CommonReportWindowAction, self).set_window_params()
        params = self.win_params.copy()
        params['title'] = self.parent.title
        params['form_url'] = self.parent.get_reporting_url()
        self.win_params = self.parent.set_report_window_params(
            params, self.request, self.context)


class CommonReportAction(objectpack.BaseAction):

    u"""Экшн, выполняющий отчёт."""

    perm_code = 'report'

    def run(self, request, context):
        out_file_url = self.parent.make_report(context)
        return download_result(out_file_url.encode('utf-8'))


class CommonReportBuilder(object):

    u"""Класс построителя отчётов для CommonReportPack."""

    # доступные форматы
    _available_extensions = ['.xls', '.xlsx', '.docx']

    # формат по-умолчанию
    extension = '.xls'

    @property
    def check_extension(self):
        u"""Проверка допустимости формата."""
        return self.extension in self._available_extensions

    def set_file_and_url(self):
        u"""Возвращает кортеж из абс пути готового отчета и url для скачивания.
        Не требует переопределения
        """
        base_name = str(uuid.uuid4())[0:16] + self.extension
        out_file_path = get_path(base_name)
        out_file_url = get_url(base_name)

        return (out_file_path, out_file_url)

    def set_up_report(self, template_path):
        u"""Настройка формата отчёта."""
        assert template_path, (
            u'%s Template for report is not defined!' % self.extension
        )
        if self.extension == '.xls':
            report = SpreadsheetReport(
                template_path, wrapper=DocumentXLS, type=FileConverter.XLS
            )
        elif self.extension == '.xlsx':
            report = SpreadsheetReport(template_path)
        elif self.extension == '.docx':
            report = DocumentReport(template_path)
        else:
            raise Exception('Unknown template extension')

        return report

    def build(self, report, params):
        u"""Метод, который делает отчёт.

        :param report: инстанс отчёта
        :type report: simple_report.report.SpreadsheetReport
        """
        assert isinstance(report, SpreadsheetReport)

        # получение абс пути отчета
        out_file, out_url = self.set_file_and_url()

        # Если пытаемся сохранить файл в несуществующую директорию,
        # то попробуем её предварительно создать
        cat = os.path.split(out_file)[0]
        if not os.path.exists(cat):
            try:
                os.makedirs(cat)
            except:
                raise ApplicationLogicException(
                    "Can't access to specified result file path"
                )

        # построение отчёта
        if isinstance(report, DocumentReport):
            report.build(out_file, params)
        else:
            report.build(out_file)

        return out_url


class CommonReportPack(BaseReportPack):

    u"""Базовый пак, реализующий генерацию отчетов, используя simple_report."""

    # доступные форматы
    _available_extensions = ['.xls', '.xlsx', '.docx']
    # формат по-умолчанию
    extension = '.xls'

    # путь где лежит файл шаблона отчета
    template_file_path = None
    # будет искать шаблон report.{extension} в директори отчета
    # template_file_path = None
    # custom_report.{extension} по абсолютному пути
    # template_file_path = '/tmp/some/custom_report.{extension}'
    # по относительному ./templates в директории отчета
    # template_file_path = './templates/custom_report.{extension}'
    # report.{extension} в ./templates/somedir/
    # template_file_path = './templates/somedir/'

    # свой внутренний построитель отчёта
    _main_builder_class = CommonReportBuilder

    def __init__(self):
        super(CommonReportPack, self).__init__()

        self.report_window_action = CommonReportWindowAction()
        self.report_action = CommonReportAction()
        self.actions.extend([
            self.report_window_action,
            self.report_action,
        ])

        self.main_builder = self._main_builder_class()

    def get_reporting_url(self):
        u"""Отдаёт адрес форме, куда передавать данные для обработки."""
        return self.report_action.get_absolute_url()

    def get_win_url(self):
        return self.report_window_action.get_absolute_url()

    def get_template(self, default_base_name='report'):
        u"""Возвращает путь к шаблону отчета.

        :param default_base_name: базовое имя шаблона,
            если определить не удалось
        :type default_base_name: str
        :returns: str - полный путь к шаблону
        """

        if (self.template_file_path is not None and
                os.path.isabs(self.template_file_path)):
            return self.template_file_path

        report_file_path = sys.modules[self.__module__].__file__
        assert __file__ != report_file_path, (
            "set_template unavailable from base class")

        report_dir = os.path.dirname(report_file_path)

        rel_sub_path = ''
        if (self.template_file_path is not None and
                self.template_file_path.startswith('./')):
            rel_sub_path = os.path.relpath(
                os.path.dirname(self.template_file_path))

        if self.template_file_path is None:
            base_name, _ = os.path.splitext(os.path.basename(report_dir))
        else:
            base_name, _ = os.path.splitext(
                os.path.basename(self.template_file_path))
            if not base_name:
                base_name = default_base_name

        auto_report_name = '{0}{1}{2}'.format(
            base_name, os.path.extsep, self.extension.strip('.'))
        auto_report_path = os.path.join(
            report_dir, rel_sub_path, auto_report_name)

        assert os.path.isfile(auto_report_path), (
            u"Report template '{0}' not found at {1}".format(
                auto_report_name, auto_report_path))
        return auto_report_path

    def set_up_report(self):
        u"""Настройка формата отчёта."""

        template_path = self.get_template()

        if self.extension == '.xls':
            report = SpreadsheetReport(
                template_path, wrapper=DocumentXLS, type=FileConverter.XLS)
        elif self.extension == '.xlsx':
            report = SpreadsheetReport(template_path)
        elif self.extension == '.docx':
            report = DocumentReport(template_path)
        else:
            raise Exception('Unknown template extension')

        return report

    def set_file_and_url(self):
        """
        Возвращает кортеж из абс. пути готового отчёта и ссылки для скачивания
        Не требует переопределения
        """

        base_name = str(uuid.uuid4())[0:16] + self.extension
        out_file_path = get_path(base_name)
        out_file_url = get_url(base_name)

        return (out_file_path, out_file_url)

    @staticmethod
    def context2dict(context):
        u"""преобразование контекста в словарь."""
        result = {}
        for key, value in context.__dict__.iteritems():
            try:
                if callable(value):
                    value = value()
                result[key] = value
            except TypeError:
                continue

        return result

    def check_report_params(self, context):
        u"""Проверка передаваемых параметров для формирования отчёта."""
        pass

    def init_provider(self, context):
        u"""Инициация провайдера - извлечение данных из контекста."""
        params = self.context2dict(context)
        self.data_provider.init(**params)

    def make_report(self, context):
        u"""Метод, реализующий отчёт.

        Не требует переопределения.
        """
        self.check_report_params(context)
        # получение абс. пути отчёта
        out_file, out_url = self.set_file_and_url()
        # настройка формата отчета
        report = self.set_up_report()

        # создание провайдера
        self.create_provider(context)
        if self.data_provider is None:
            self.data_provider = self.data_provider_class()

        self.init_provider(context)
        self.data_provider.load_data()

        # создание билдера, который будет строить кастомный отчёт
        self.create_builder(context, report=report)

        params = self.report_builder.build()

        # Если пытаемся сохранить файл в несуществующую директорию,
        # то попробуем её предварительно создать
        cat = os.path.dirname(out_file)
        if not os.path.exists(cat):
            try:
                os.makedirs(cat)
            except OSError:
                raise ApplicationLogicException(
                    "Can't access to specified result file path "
                )

        # построение отчёта
        if isinstance(report, DocumentReport):
            report.build(out_file, params)
        else:
            report.build(out_file)

        return out_url

    def set_report_window_params(self, params, request, context):
        u"""Дополнение параметров окна отчёта."""
        if not self.main_builder.check_extension:
            raise ApplicationLogicException(u'Расширение указано неверно!')
        params['extension'] = self.main_builder.extension
        return params

    def create_report_window(self, request, context):
        return self.report_window()

    def extend_menu(self, menu):
        u"""Размещение в меню."""
        pass
